import base64
import subprocess
from dataclasses import dataclass, field
from http.client import OK
from pathlib import Path

import requests

from .utils import cache_dir


@dataclass
class Icon:
    name: str
    cache_dir: Path = field(default_factory=cache_dir)
    size: int = 20

    def __str__(self) -> str:
        return self.png_base64()

    @property
    def png(self) -> Path:
        if not (f := self.cache_dir / f"{self.name}.png").exists():
            self.fetch(f)
        return f

    def fetch(self, f: Path) -> None:
        raise NotImplementedError()

    def resize(self, size: int = 0) -> Path:
        size = size or self.size
        if not (f := self.cache_dir / f"{self.name}.{size}.png").exists():
            subprocess.run(
                f"sips -Z {size} {self.png} --out {f}",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        return f

    def png_crush(self, size: int = 0) -> Path:
        size = size or self.size
        pngcrush = "/Applications/Xcode.app/Contents/Developer/usr/bin/pngcrush"

        if not (f := self.cache_dir / f"{self.name}.{size}.crush.png").exists():
            subprocess.run(
                f"{pngcrush} -brute {self.resize(size)} {f}",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        return f

    def png_base64(self, size: int = 0) -> str:
        return (
            (
                base64.encodebytes(self.png_crush(size or self.size).read_bytes())
                .decode()
                .replace("\n", "")
            )
            if self.png.exists()
            else ""
        )


@dataclass
class UrlIcon(Icon):
    url: str = ""

    def fetch(self, f: Path):
        response = requests.get(self.url)
        if OK == response.status_code:
            f.write_bytes(response.content)


@dataclass
class ServiceIcon(Icon):
    def fetch(self, f: Path):
        url = f"https://raw.githubusercontent.com/walkxcode/dashboard-icons/main/png/{self.name}.png"
        response = requests.get(url)
        if OK == response.status_code:
            f.write_bytes(response.content)


@dataclass
class IcnsIcon(Icon):
    icns_size: int = 32

    def fetch(self, f: Path):
        for apps in ["/System/Applications", "/Applications"]:
            if (app := (Path(apps) / f"{self.name}.app")).exists():
                # find icon
                icns = subprocess.check_output(
                    f"/usr/libexec/PlistBuddy  -c 'Print :CFBundleIconFile' {app}/Contents/Info.plist",
                    shell=True,
                    encoding="utf8",
                    stderr=subprocess.DEVNULL,
                ).strip()

                # extract icon
                subprocess.run(
                    f"icns2png --size {self.icns_size} -x {app}/Contents/Resources/{icns}.icns --out {f.parent}",
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                # reanme icon
                (
                    f.parent
                    / f"{icns}_{self.icns_size}x{self.icns_size}x{self.icns_size}.png"
                ).rename(f)
                break



from __future__ import annotations

import json
import logging
import sys
from dataclasses import dataclass, field
from inspect import getfile
from pathlib import Path
from typing import ClassVar, Iterable, Type, TypeVar, cast, get_type_hints

from pyxbar.types import Renderable, RenderableGenerator
from pyxbar.utils import cache_dir

logger = logging.getLogger()
logging.basicConfig(level=logging.DEBUG, format="=====> %(message)s")


@dataclass
class Config(Renderable):
    DEBUG: bool = False
    MONO_FONT: str = "Andale Mono"
    CACHE_DIR: Path = field(default_factory=cache_dir, repr=False)

    _errors: list[str] = field(default_factory=list, init=False, repr=False)
    _warnings: list[str] = field(default_factory=list, init=False, repr=False)

    prefix: ClassVar[str] = "VAR_"

    def __post_init__(self):
        for k in self.config_fields():
            v = getattr(self, k)
            if isinstance(v, Path):
                setattr(self, k, v.expanduser().resolve())
            if self.DEBUG:
                logger.debug(f"{k}: {getattr(self, k)}")

    @classmethod
    def config_fields(cls) -> Iterable[str]:
        return [k for k in cls.__dataclass_fields__ if k.isupper()]

    @classmethod
    def config_path(cls) -> Path:
        return Path(f"{getfile(cls)}.vars.json")

    @classmethod
    def from_config_file(cls):
        loaded = json.loads(cls.config_path().read_text() if cls.config_path().exists() else "{}")
        return cls.from_config_dict(loaded)

    @classmethod
    def from_config_dict(cls, config_dict: dict[str, str]):
        hints = get_type_hints(cls)
        return cls(
            **{
                k: hints[k](config_dict[f"VAR_{k}"])
                for k in cls.config_fields()
                if f"VAR_{k}" in config_dict
            }
        )

    @classmethod
    def get_config(cls):
        return cls.from_config_file()

    def as_config_dict(self) -> dict[str, str]:
        return {
            f"{self.prefix}{k}": str(getattr(self, k)) for k in self.config_fields()
        }

    def save(self):
        self.config_path().write_text(json.dumps(self.as_config_dict(), indent=4))

    def render(self, depth: int = 0) -> RenderableGenerator:
        from pyxbar import Divider, MenuItem

        depth_prefix = f"{'--' * depth}"
        for title, color, preifx, messages in (
            ("errors", "red", "❌", self._errors),
            ("warnings", "yellow", "⚠️", self._warnings),
        ):
            if messages:
                yield from Divider().render(depth)
                yield from MenuItem(title, color=color).render(depth)
                for msg in sorted(messages):
                    yield from MenuItem(f"{depth_prefix} {preifx} {msg}").render(depth)

        if self.DEBUG:
            yield from MenuItem(f"🐍 {sys.version}").render()
            yield from MenuItem("Vars").with_submenu(
                MenuItem(f"{k}: {getattr(self, k)}") for k in self.__dataclass_fields__
            ).render(depth)
            # yield Cmd(f"📝 Edit Vars", f"open '{__file__}.vars.json'", depth=2)

    def error(self, msg: str):
        if msg not in self._errors:
            self._errors.append(msg)

    def warn(self, msg: str):
        if msg not in self._warnings:
            self._warnings.append(msg)


ConfigT = TypeVar("ConfigT", bound=Config)

CONFIG_CLS: Type[Config]
CONFIG: Config


def get_config(config_cls: Type[ConfigT] = Config) -> ConfigT:
    global CONFIG_CLS, CONFIG

    if config_cls is globals().get("CONFIG_CLS") is None:
        raise RuntimeError("CONFIG_CLS is not set")

    if config_cls:
        CONFIG_CLS = config_cls  # type: ignore

    if "CONFIG" in globals() and isinstance(CONFIG, CONFIG_CLS):
        return cast(ConfigT, CONFIG)

    globals()["CONFIG"] = CONFIG_CLS.get_config()

    return get_config(config_cls=config_cls)
